/****************************************************************************
 ** hw_sudp.c ***************************************************************
 ****************************************************************************
 *
 * receive input via UDP.
 * 
 * This file is based on hw_udp.c
 *
 * Copyright (C) 2007 Joni Valtanen <jvaltane@kapsi.fi>
 * Copyright (C) 2002 Jim Paris <jim@jtan.com>
 *
 * Distribute under GPL version 2 or later.
 *
 * Received UDP packets consist of some number of LE 16-bit integers.
 * Integer simply specify button number what is pushed.
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/fcntl.h>
#include <netinet/in.h>
#include <errno.h>

#include "hardware.h"
#include "ir_remote.h"
#include "lircd.h"
#include "receive.h"
#include "transmit.h"
#include "hw_default.h"

static int zerofd;       /* /dev/zero */
static int sockfd;       /* the socket */
static lirc_t data;

int hwsudp_decode(struct ir_remote *remote,
	       ir_code *prep,ir_code *codep,ir_code *postp,
		 int *repeat_flagp,lirc_t *remaining_gapp);

int hwsudp_init()
{
	int port;
	struct sockaddr_in addr;

	logprintf(LOG_INFO,"Initializing UDP: %s",hw.device);
	
	init_rec_buffer();
	
	port=atoi(hw.device);
	if(port==0) {
		logprintf(LOG_ERR,"invalid port: %s",hw.device);
		return 0;
	}

	/* hw.fd needs to point somewhere when we have extra data */
	if((zerofd=open("/dev/zero",O_RDONLY))<0) {
		logprintf(LOG_ERR,"can't open /dev/zero: %s",
			  strerror(errno));
		return 0;
	}

	if((sockfd=socket(AF_INET,SOCK_DGRAM,0))<0) {
		logprintf(LOG_ERR,"error creating socket: %s",
			  strerror(errno));
		close(zerofd);
		return 0;
	}
	
	addr.sin_family      = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port        = htons(port);
	
	if(bind(sockfd,(struct sockaddr *)&addr,sizeof(addr))<0) {
		logprintf(LOG_ERR,"can't bind socket to port %d: %s",
			  port,strerror(errno));
		close(sockfd);
		close(zerofd);
		return 0;
	}
	
	logprintf(LOG_INFO,"Listening on port %d/udp",port);

	hw.fd=sockfd;

	return(1);
}

int hwsudp_deinit(void)
{
	close(sockfd);
	close(zerofd);
	hw.fd=-1;
	return(1);
}

char *hwdecode_all(struct ir_remote *remotes)
{
  struct ir_remote *remote;
  static char message[PACKET_SIZE+1];
  ir_code pre,code,post;
  struct ir_ncode *ncode;
  int repeat_flag;
  ir_code toggle_bit_mask_state;
  lirc_t remaining_gap;
  struct ir_remote *scan;
  struct ir_ncode *scan_ncode;
  struct ir_ncode *codes, *found;
  int found_code, have_code;
  size_t len = 0;  
  have_code = 0;
  found_code = 0;
  found = NULL;

  /* use remotes carefully, it may be changed on SIGHUP */
  remote=remotes;
  while(remote)
    {
        codes=remote->codes;
        if(codes!=NULL)
        {
	  while(codes->name!=NULL)
          {
	    logprintf(LOG_INFO, "code name: %s\n",codes->name);
            remote->last_code = codes;
	    
	    ir_code next_code;
	    if(codes->next!=NULL && codes->current!=NULL)
	      {
		next_code=codes->current->code;
	      }
	    else
	      {
		next_code=codes->code;
	      }

	      logprintf(LOG_INFO, "nexcode: %d\n",next_code);
	      logprintf(LOG_INFO, "data: %d\n",data);

	      if(data == next_code)
		{
 	          found=codes;
		  goto outerlimits;
		}
	      else
		{
		  codes->current=NULL;
		}
	      codes++;
	  }
	}
      remote = remote->next;
    }

 outerlimits:
#ifdef __GLIBC__
  /* It seems you can't print 64-bit longs on glibc */

  if( found==NULL )
    return NULL;
  if(remote == NULL)
    return NULL;
  if( remote->last_code == NULL )
    return NULL;

  if( found->current!=NULL )
  {
    return(NULL);
  }

  len=snprintf(message,PACKET_SIZE+1,"%08lx%08lx %02x %s %s\n",
	       (unsigned long)(found->code>>32),
	       (unsigned long)(found->code&0xFFFFFFFF),
	       0,
	       remote->last_code->name,
	       remote->name);

#else
  len=snprintf(message,PACKET_SIZE+1,"%016llx %02x %s %s\n
",
	       found->code,
	       remote->reps-(found->next ? 1:0),
	       remote->last_code->name,
	       remote->name);
#endif

  if(len>=PACKET_SIZE+1)
    {
      logprintf(LOG_ERR,"message buffer overflow");
      return(NULL);
    }
  logprintf(LOG_INFO, "message: %s\n",message);

  return(message);

}

char *hwsudp_rec(struct ir_remote *remotes)
{
  char *retval = NULL; // = "FF";
	if(!clear_rec_buffer()) 
	       return(NULL);
	retval = hwdecode_all(remotes);
	
	return(retval);
}

int hwsudp_decode(struct ir_remote *remote,
	       ir_code *prep,ir_code *codep,ir_code *postp,
	       int *repeat_flagp,lirc_t *remaining_gapp)
{
  /*  int retval = receive_decode(remote,prep,codep,postp,
  	    repeat_flagp,remaining_gapp);

	logprintf(LOG_INFO, "decode retval: %d\n",retval);
	return(retval);
  */

  if (remote == NULL) 
    return(NULL);
  if( remote->last_code == NULL)
    return(NULL);

  *prep=remote->pre_data;
  *codep=remote->last_code->code;
  *postp=remote->post_data;
  *repeat_flagp=1;
  *remaining_gapp=1;

  return(1);
}

lirc_t hwsudp_readdata(lirc_t timeout)
{
	static u_int8_t buffer[8192];
	static int buflen=0;
	static int bufptr=0;
	u_int8_t packed[2];
	u_int32_t tmp;
	fd_set rfd;
	struct timeval tv;

	/* Assume buffer is empty; LIRC should select on the socket */
	hw.fd=sockfd;

	/* If buffer is empty, get data into it */
	if((bufptr+2)>buflen) 
	{
		FD_ZERO(&rfd);
		FD_SET(sockfd,&rfd);
		tv.tv_sec=0;
		tv.tv_usec=timeout;
		if(select(sockfd+1,&rfd,NULL,NULL,&tv)!=1) 
			return 0;
		if((buflen=recv(sockfd,&buffer,sizeof(buffer),0))<0)
		{
			logprintf(LOG_INFO,"Error reading from UDP socket");
			return 0;
		}
		if(buflen&1) 
			buflen--;
		if(buflen==0)
			return 0;
		bufptr=0;
	}

	/* Read as 2 bytes to avoid endian-ness issues */
	packed[0]=buffer[bufptr++];
	packed[1]=buffer[bufptr++];

	data = (lirc_t)packed[1];
	data = data + (((lirc_t)packed[0])<<8);

	/* If our buffer still has data, give LIRC /dev/zero to select on */
	if((bufptr+2)<=buflen)
		hw.fd=zerofd;

	logprintf(LOG_INFO, "packed[0]: %d packed[1]: %d data: %d\n", 
		  packed[0], packed[1], data);
	return(data);
}

struct hardware hw_sudp=
{
	"9876",	    	    /* "device" (port) */
	-1,                 /* fd (socket) */
	LIRC_CAN_REC_RAW, /* features */
	0,                  /* send_mode */
	LIRC_MODE_RAW,    /* rec_mode */
	0,                  /* code_length */
	hwsudp_init,	    /* init_func */
	NULL,		    /* config_func */
	hwsudp_deinit,         /* deinit_func */
	NULL,		    /* send_func */
	hwsudp_rec,            /* rec_func */
	hwsudp_decode,         /* decode_func */
	NULL,               /* ioctl_func */
	hwsudp_readdata,       /* readdata */
	"sudp"
};
